from durable_dot_dict.dot import Dot
from durable_dot_dict.dotdict import DotDict

class X:
    x = 100

def test_dot():
    dotdict = DotDict({"a": {"b": 1, "c": [{"d": 2, "e": X()}]}})
    dot = Dot(dotdict)
    # assert dot.a.c == [{"d": 2, "e": X()}]
    print(dot['a'])
    print(dot.a.c[0])
    print(dot.a.c[0].d)
    print(dot.a.c[0].e.x)

    dot = Dot()
    dot.a = 1
    dot.b = {}
    dot.b.a = 1
    print(dot)



import pytest

from durable_dot_dict.collection import first
from durable_dot_dict.dotdict import DotDict


def test_init():

    # No SqlDB style
    assert DotDict(
        {
            'b': {'a': None},
            'c': {'a': 1},
            'd': {'a': ''},
            'e': {'a': {'b': 1}},
            'f': {'a': [{'a': 1}, 1]},
            'list': [1],
            'set': set()}
    ) == {'b': {'a': None}, 'c': {'a': 1}, 'd': {'a': ''}, 'e': {'a': {'b': 1}}, 'f': {'a': [{'a': 1}, 1]}, 'list': [1],
          'set': set()}

    # FlatDB style
    assert DotDict() << {
        "b.a": None,
        'c.a': 1,
        'd.a': "",
        'e.a': {"b": 1},
        'f.a': [{"a": 1}, 1],
        "list": [1],
        "set": set()
    } == {'b': {'a': None}, 'c': {'a': 1}, 'd': {'a': ''}, 'e': {'a': {'b': 1}}, 'f': {'a': [{'a': 1}, 1]}, 'list': [1],
          'set': set()}

    # FlatDB style
    assert DotDict() << [
        ("b.a",  None),
        ('c.a', 1),
        ('d.a', ""),
        ('e.a', {"b": 1}),
        ('f.a', [{"a": 1}, 1]),
        ("list", [1]),
        ("set", set())
    ] == {'b': {'a': None}, 'c': {'a': 1}, 'd': {'a': ''}, 'e': {'a': {'b': 1}}, 'f': {'a': [{'a': 1}, 1]}, 'list': [1],
          'set': set()}


def test_flat():
    data = DotDict({
        "b.a": None,
        'c.a': 1,
        'd.a': "",
        'e.a': {"b": 1},
        'f.a': [{"a": 1}, 1],
        "list": [1],
        "set": set()
    }) << [
               ('list[1]', 2)
           ]
    assert data.flat() == {'b.a': None, 'c.a': 1, 'd.a': '', 'e.a.b': 1, 'f.a': [{'a': 1}, 1], 'list': [1, 2],
                           "set": set()}


def test_none():
    data = DotDict({
        "b": None,
        'c': 1,
        'd': "",
        'e': {},
        'f': []
    })

    assert bool(data)
    assert data.empty('a')
    assert data.empty('a.c')
    assert data.empty('b')
    assert not data.empty('c')
    assert data.empty('d')
    assert data.empty('e')
    assert data.empty('f')

    assert not bool(DotDict())


def test_first():
    a1 = {}
    a2 = {"b": 1}
    data = DotDict({})
    data['a'] = first(lambda: a1["a"], lambda: a2["b"])
    assert data == {'a': 1}


def test_map_key_value():
    data = DotDict()
    data << [
        ('a', 1),
        ('b.c', 2)
    ]
    assert data == {'a': 1, 'b': {'c': 2}}

    assert DotDict() << [('a', 1), ('b.c', 2)] == {'a': 1, 'b': {'c': 2}}

    dot = DotDict() << [('a', 1), ('b.c', 2)]
    assert dot >> [
        ("a", 'a'),
        ("b_c", 'b.c')
    ] == {'a': 1, 'b_c': 2}


def test_dotdict_deep_set():
    data = DotDict({})
    data['data.bookings[0].services[0].details.airline'] = 'Air France'
    assert data == {'data': {'bookings': [{'services': [{'details': {'airline': 'Air France'}}]}]}}

    data = DotDict({})
    data['data[0][0].detail'] = 'Air France'
    assert data == {'data': [[{'detail': 'Air France'}]]}

    data = DotDict({})
    data['data[0][0]'] = 'Air France'
    assert data == {'data': [['Air France']]}

    data = DotDict({})
    data['data.list[0][0]'] = {1}
    assert data == {'data': {"list": [[{1}]]}}


def test_embedded_dotdict():
    d = DotDict({
        "data": {
            "list": [[1, 2, 3], [4, 5, 6], [7, 8, 9]],
            "bookings": [
                {
                    "services": [
                        {
                            "details": {
                                "airline": "Air France",
                            }
                        }
                    ]
                }
            ]
        }
    })

    assert d['data.bookings[0].services[0].details.airline'] == 'Air France'
    assert 'data.bookings[1].services[0]' not in d
    assert 'data.bookings[0].services[0]' in d
    assert 'data.bookings[0].services' in d
    assert 'data.list[0][1]' in d
    assert d['data.list[0][1]'] == 2


def test_dotdict_spread():
    d1 = DotDict({
        "a": "1"
    })
    d2 = DotDict({
        "b": "2"
    })

    print({**d1, **d2})


def test_dotdict_set_as_dotdict():
    d = DotDict({
        "a": "1"
    })

    with pytest.raises(TypeError):
        cd = DotDict(d)

    d = DotDict({"a": 1})
    d['a'] = DotDict({"b": {"c": [2, 1]}})
    assert isinstance(d['a'], dict)
    assert d['a.b.c.0'] == 2
    assert d['a']['b']['c'][0] == 2


# Define tests for DotDict functionality
def test_dotdict_set_get_delete_check():
    d = {
        "a": {"b": ["c", 0]},
        "a1": {"b": ["c", 0]},
        "a2": {"b": ["c", 0]},
        "a3": {"b": ["c", 0, {"c"}, ("x", 10)]}
    }

    cd = DotDict(d)

    # Test setting and getting using unified path notation
    cd['a3.b[3]'] = "new_value"
    assert cd['a3.b[3]'] == "new_value", "Failed to set or get value using unified path notation"
    assert cd['a3']['b'][3] == "new_value", "Failed to set or get value using unified path notation"

    # Test setting and getting using attribute-style notation
    cd['a3.b'].append("another_value")
    assert cd['a3.b[4]'] == "another_value"

    # Test deletion using unified path notation

    # assert cd.a3.b == DotDict(['c', 0, {'c'}, 'new_value', 'another_value'])
    assert cd['a3.b'] == ['c', 0, {'c'}, 'new_value', 'another_value']

    del cd['a3.b[2]']
    assert cd['a3.b'] == ['c', 0, 'new_value', 'another_value']
    print(cd['a3.b'])
    # Test checking existence using unified path notation
    assert 'a3.b[2]' in cd, "Membership test for 'a3.b[1]' failed"
    assert 'a3.b[10]' not in cd, "Non-membership test for 'a3.b[10]' failed"


def test_dotdict_get():
    d = {
        "a": {"b": ["c", 0]},
        "b": [{"$c": "here"}],
        "A": {"a b c": 1}
    }

    data = DotDict(d)
    with pytest.raises(TypeError):
        x = data['a.b[]']

    assert data["b.0['$c']"] == 'here'
    assert data['A["a b c"]'] == 1


def test_dotdict_get_default():
    d = {}

    data = DotDict(d)
    x = data.get('xx.xx', "default")
    assert x == 'default'
    x = data.get('xx.xx', None)
    assert x is None


def test_dotdict_has():
    d = {
        "a": {"b": ["c", 0]},
    }

    data = DotDict(d)
    assert 'a.b' in data
    assert 'a.d' not in data
    assert 'a[1].d' not in data
    assert 'a[]' not in data
    assert 'a.b[1]' in data


def test_dotdict_set():
    data = DotDict({})

    # Test setting and getting using unified path notation
    data['a.b[0]'] = 123
    assert data['a.b[0]'] == 123
    data['a.b'].append("1")
    data['a.b'].append("2")
    assert data['a.b'] == [123, "1", "2"]
    with pytest.raises(TypeError):
        data['a.b[0].c'] = 123


def test_equal():
    x = DotDict({'id': '1', 'active': True, 'metadata': {
        'time': {'insert': '2025-01-10T17:13:28.620880+00:00', 'create': '2004-01-12T17:13:28.620880+00:00',
                 'update': '2025-03-20T10:53:41.924819+00:00'}}, 'operation': {'new': False, 'update': False},
                 'ids': []})
    y = DotDict({'id': '1', 'active': True, 'metadata': {
        'time': {'insert': '2025-01-10T17:13:28.620880+00:00', 'create': '2004-01-12T17:13:28.620880+00:00',
                 'update': '2025-03-20T10:53:41.924819+00:00'}}, 'operation': {'new': False, 'update': False},
                 'ids': []})
    assert x == y


# Define tests for invalid data paths
def test_invalid_paths_handling():
    d = {
        "a": {"b": ["c", 0]},
        "a3": {"b": ["c", 0, {"c"}, ("x", 10)]}
    }

    cd = DotDict(d)

    invalid_paths = [
        ".leadingDot",  # Leading dot
        "trailingDot.",  # Trailing dot
        "a..b",  # Multiple consecutive dots
        "a[\"b-\"c0\"]",  # Invalid string key in bracket
        "a[\"b-\"c0\"][0]",
        "a[\"b-\"c0\"].ala.",
        "",
        "a[\"b-\"c0\".ala",  # Unclosed bracket
    ]

    for path in invalid_paths:
        with pytest.raises((ValueError, KeyError)):
            _ = cd[path]


# Run all tests when executed via pytest
if __name__ == "__main__":
    pytest.main()



import copy
import json
from typing import Union, List, Tuple, Any, Optional, Dict, Set
from collections.abc import MutableMapping
import dotdict_parser


class DotDict(MutableMapping):
    def __init__(self, dictionary: Optional[Union[dict | list]] = None):
        if dictionary is None:
            dictionary = {}
        if not isinstance(dictionary, (dict, list)):
            raise TypeError(f"Expected dictionary or list as DotDict. Got {type(dictionary)}.")
        self.root = dictionary

    def _set_path_value(self, path, value):
        """
        Walks through `root` (which should be a dict or list at the top level),
        creating intermediate dicts/lists as needed so that each element in `path`
        is valid. The final element of `path` will be set to `value`.
        """
        node = self.root

        for i in range(len(path) - 1):
            key = path[i]
            next_key = path[i + 1]

            if isinstance(key, str):
                # Ensure `node` is a dict if we are using a string key
                if not isinstance(node, dict):
                    raise TypeError(f"Cannot use string key on non-dict: {node}")
                # If key doesn't exist, create either a dict or list based on the next key
                if key not in node:
                    node[key] = [] if isinstance(next_key, int) else {}
                node = node[key]

            elif isinstance(key, int):
                # Ensure `node` is a list if we are using an integer key
                if not isinstance(node, list):
                    raise TypeError(f"Cannot use integer key on non-list: {node}")
                # Expand the list if needed
                while len(node) <= key:
                    node.append(None)
                # If there's nothing at node[key], create either a dict or list for the next step
                if node[key] is None:
                    node[key] = [] if isinstance(next_key, int) else {}
                node = node[key]

            else:
                raise TypeError(f"Keys must be str or int, got {type(key)}")

        # Convert DotDict to dict
        if isinstance(value, DotDict):
            value = value.to_dict()

        # Handle the last key in the path and set `value`
        last_key = path[-1]
        if isinstance(last_key, str):
            if not isinstance(node, dict):
                raise TypeError(f"Cannot assign string-key '{last_key}' to non-dict: {node}")
            node[last_key] = value
        elif isinstance(last_key, int):
            if not isinstance(node, list):
                raise TypeError(f"Cannot assign integer-key '{last_key}' to non-list: {node}")
            while len(node) <= last_key:
                node.append(None)
            node[last_key] = value
        else:
            raise TypeError(f"Keys must be str or int, got {type(last_key)}")

    def _set_reference(self, path, key):
        data = self.root
        print(path)
        for item_no, item in enumerate(path):
            print(item)
            if isinstance(item, int):
                if not isinstance(data, list):
                    data = []
                data[item] = []
            elif isinstance(item, str):
                if item not in data:
                    data[item] = {}
                data = data[item]
            else:
                raise KeyError(f"Only string keys are allowed. Got {item} of type {type(item)}.")
        return data

    def _has_reference(self, keys) -> bool:
        data = self.root
        last = len(keys) - 1
        for pos, key in enumerate(keys):
            if isinstance(key, int):
                # Is int but dat is not list
                if not isinstance(data, list):
                    return False
                # Is last so check number of items
                if pos == last:
                    return len(data) > key
                else:
                    # Not last so check if key not out of range
                    if len(data) <= key:
                        return False

            elif key not in data:
                return False
            data = data[key]
        return True

    def _reference(self, keys):
        data = self.root
        for key in keys:
            data = data[key]
        return data

    @staticmethod
    def _path_key(keys):
        path = keys[:-1]
        key = keys[-1]
        return path, key

    def get(self, key, *args):
        try:
            keys = dotdict_parser.parse_unified_path(key)
            return self._reference(keys)
        except (ValueError, KeyError) as e:
            if args:
                return args[0]
            raise KeyError(f"Could not get DotDict value for {key}. Default value: {args}. Details: {str(e)}")

    def copy(self):
        return DotDict(self.root.copy())

    def deep_copy(self):
        return DotDict(copy.deepcopy(self.root))

    def to_dict(self) -> dict:
        return self.root

    def to_json(self, default=None, cls=None):
        return json.dumps(self.root, default=default, cls=cls)

    @staticmethod
    def as_list(data: List[dict]) -> List['DotDict']:
        return list(map(DotDict, data))

    def get_or_none(self, key: str) -> Optional[Any]:
        return self.get(key, None)

    def empty(self, key) -> bool:
        result = self.get(key, None)

        if isinstance(result, (bool, int, float)):
            return False

        return result is None or not bool(result)

    def flat(self):
        return dotdict_parser.flatten(self.root)

    def map(self, right: Union['DotDict', dict]) -> 'Mapper':
        return Mapper(self, right)

    def __contains__(self, item):
        keys = dotdict_parser.parse_unified_path(item)
        return self._has_reference(keys)

    def __getitem__(self, item):
        if isinstance(item, int):
            return self.root[item]
        return self.get(item)

    def __setitem__(self, key, value):
        keys = dotdict_parser.parse_unified_path(key)
        self._set_path_value(keys, value)

    def __delitem__(self, key):
        if isinstance(key, int):
            del self.root[key]
        else:
            keys = dotdict_parser.parse_unified_path(key)
            path, key = self._path_key(keys)
            data = self._reference(path)
            del data[key]

    def __repr__(self):
        return f'{self.__class__}({self.root})'

    def __str__(self):
        return self.root.__str__()

    def __hash__(self):
        return self.root.__hash__()

    def __len__(self):
        return self.root.__len__()

    def __getstate__(self):
        return self.root

    def __setstate__(self, state):
        self.root = state

    def __iter__(self):
        # Return an iterator over the keys
        return self.root.__iter__()

    def __eq__(self, other):
        if isinstance(other, DotDict):
            return other.to_dict() == self.to_dict()
        elif isinstance(other, dict):
            return other == self.to_dict()
        else:
            return False

    def __lshift__(self, list_of_kv) -> 'DotDict':
        list_of_kv = list_of_kv.items() if isinstance(list_of_kv, dict) else list_of_kv

        for key, value in list_of_kv:
            self[key] = value

        return self

    def __rshift__(self, list_of_kv: List[Tuple[str, str]]) -> 'DotDict':
        dot = DotDict()
        for key, value in list_of_kv:
            dot[key] = self[value]
        return dot

def _data_convert(list_of_left_right):
    if isinstance(list_of_left_right, set):
        return [(item, item) for item in list_of_left_right]
    return list_of_left_right.items() if isinstance(list_of_left_right, dict) else list_of_left_right


class Mapper:

    def __init__(self, left: DotDict, right: Union[DotDict, dict]):
        self.left = left
        self.right: DotDict = DotDict(right) if isinstance(right, dict) else right

    def __lshift__(self, list_of_left_right: Union[List[Tuple[str, str]], Dict[str, str], Set[str]]) -> 'DotDict':
        list_of_left_right = _data_convert(list_of_left_right)
        for left, right in list_of_left_right:
            if right in self.right:
                self.left[left] = self.right[right]

        return self.left

    def __rshift__(self, list_of_left_right: Union[List[Tuple[str, str]], Dict[str, str], Set[str]]) -> 'DotDict':
        list_of_left_right = _data_convert(list_of_left_right)
        for left, right in list_of_left_right:
            if left in self.left:
                self.right[right] = self.left[left]
        return self.right



from typing import Any, Callable


def first(*args: Callable[[], Any]) -> Any:
    for func in args:
        try:
            return func()
        except Exception:
            continue
    raise ValueError("All provided callables failed.")


def first_not_none(*args: Callable[[], Any]) -> Any:
    for func in args:
        try:
            result = func()
            if result is None:
                continue
            return result
        except Exception:
            continue
    raise ValueError("All provided callables failed.")

